# SPDX-License-Identifier: MIT OR Apache-2.0

"""
The resources sub-module for the Avro Dictionary.

Licensed under the terms of the MIT OR Apache-2.0 License.
"""

# Import local modules.
from .dictionary import *
